/*
 * arch-ia32.h - Architecture-specific functions for ia32 (i386 and above)
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/ptrace.h>
#include <sys/user.h>

#include "util.h"
#include "ptrace.h"
#include "arch.h"


/*
 * http://www.nondot.org/gcc/i386_8c-source.html
 *
 * 00470     0 for %eax (gcc regno = 0)
 * 00471     1 for %ecx (gcc regno = 2)
 * 00472     2 for %edx (gcc regno = 1)
 * 00473     3 for %ebx (gcc regno = 3)
 * 00474     4 for %esp (gcc regno = 7)
 * 00475     5 for %ebp (gcc regno = 6)
 * 00476     6 for %esi (gcc regno = 4)
 * 00477     7 for %edi (gcc regno = 5)
 */


struct ia32_dsc {
    struct user_regs_struct regs;
    struct pt_process *proc;
};


static void ia32_stopped(void *dsc)
{
    struct ia32_dsc *d = dsc;

    if (ptrace(PTRACE_GETREGS,d->proc->pid,NULL,&d->regs) < 0) {
	perror("ptrace(PTRACE_GETREGS)");
	exit(1);
    }
}


static void ia32_resume(void *dsc)
{
    const struct ia32_dsc *d = dsc;

    if (ptrace(PTRACE_SETREGS,d->proc->pid,NULL,&d->regs) < 0) {
	perror("ptrace(PTRACE_SETREGS)");
	exit(1);
    }
}


static void ia32_set_reg(void *dsc,int reg,arch_reg_t v)
{
    struct ia32_dsc *d = dsc;

    switch (reg) {
	case 0:
	    d->regs.eax = v;
	    break;
	case 1:
	    d->regs.ecx = v;
	    break;
	case 2:
	    d->regs.edx = v;
	    break;
	case 3:
	    d->regs.ebx = v;
	    break;
	case 4:
	    d->regs.esp = v;
	    break;
	case 5:
	    d->regs.ebp = v;
	    break;
	case 6:
	    d->regs.esi = v;
	    break;
	case 7:
	    d->regs.edi = v;
	    break;
	default:
	    abort();
    }
}


static arch_reg_t ia32_get_reg(void *dsc,int reg)
{
    const struct ia32_dsc *d = dsc;

    switch (reg) {
	case 0:
	    return d->regs.eax;
	case 1:
	    return d->regs.ecx;
	case 2:
	    return d->regs.edx;
	case 3:
	    return d->regs.ebx;
	case 4:
	    return d->regs.esp;
	case 5:
	    return d->regs.ebp;
	case 6:
	    return d->regs.esi;
	case 7:
	    return d->regs.edi;
	default:
	    abort();
    }
}


static void ia32_backup_pc(void *dsc)
{
    struct ia32_dsc *d = dsc;

    d->regs.eip--;
}


static arch_addr_t ia32_get_pc(void *dsc)
{
    const struct ia32_dsc *d = dsc;

    return d->regs.eip;
}


static void *ia32_get_bp(void)
{
    uint8_t *op;

    op = alloc(1);
    *op = 0xcc;	/* INT3 */   
    return op;
}


static void ia32_swap_bp(void *dsc,void *covered,arch_addr_t addr)
{
    struct ia32_dsc *d = dsc;

    pt_swap_byte(d->proc,addr,covered);
}


static void *ia32_init_arch(struct pt_process *proc)
{
    struct ia32_dsc *d;

    d = alloc_t(struct ia32_dsc);
    d->proc = proc;
    ia32_stopped(d);
    return d;
}


struct arch_ops ia32_arch_ops = {
    .init_arch = ia32_init_arch,
    .stopped = ia32_stopped,
    .resume = ia32_resume,
    .set_reg = ia32_set_reg,
    .get_reg = ia32_get_reg,
    .backup_pc = ia32_backup_pc,
    .get_pc = ia32_get_pc,
    .get_bp = ia32_get_bp,
    .swap_bp = ia32_swap_bp,
};
