/*
 * arch.h - Architecture-specific functions for ptrace and sbug
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#ifndef ARCH_H
#define ARCH_H

#include <inttypes.h>


struct pt_process; /* avoid circular includes */


/*
 * Note that just using arch_reg_t as an integer type is a bit too simplistic:
 * if we have an architecture with different register sizes, signed operations
 * depend on knowing the correct size. So we'd either need a size indicator in
 * arch_reg_t and architecture-specific arithmetic operators, or signed and
 * unsigned get (and set ?) operations. The upper layers need to track
 * signedness anyway.
 */

typedef uint32_t arch_reg_t;
typedef int32_t arch_sreg_t;	/* signed equivalent of arch_reg_t */
typedef uint32_t arch_addr_t;	/* convenience only - must be == arch_reg_t */


/*
 * init_arch	allocate and initialize the descriptor. "proc" points to the
 *		stopped process. init_arch can later on be free'd by the
 *		caller.
 * stopped	invoked whenever the process is stopped, before any other calls
 *		to architecture functions that modify registers ([gs]et_reg,
 *		{backup,get}_pc, but not {get,swap}_bp). "stopped" is _not_
 *		invoked after init_arch, at process termination, or if the
 *		process is stopped just to modify breakpoints.
 * resume	invoked whever the process resumes operation, after any other
 *		calls to architecture functions. "resume" is also invoked when
 *		the process begins execution, after init_arch.
 * set_reg	set the specified register. Writes to the process can either be
 *		performed immediately, or they can be deferred until "resume".
 *		In either case, a subsequent call to "get_reg" must return the
 *		modified value.
 * get_reg	read the specified register and return its value.
 * backup_pc	adjust the program counter after a breakpoint exception.
 * get_pc	retrieve the current program counter. When handling a
 *		breakpoint, "get_pc" is called after "backup_pc".
 * get_bp	allocate space for the breakpoint opcode, and initialize that
 *		space to the opcode value. The space can later on be free'd by
 *		the caller.
 * swap_bp	swap the opcode at "addr" with the opcode stored at "covered".
 */


struct arch_ops {
    void *(*init_arch)(struct pt_process *proc);
    void (*stopped)(void *dsc);
    void (*resume)(void *dsc);
    void (*set_reg)(void *dsc,int reg,arch_reg_t v);
    arch_reg_t (*get_reg)(void *dsc,int reg);
    void (*backup_pc)(void *dsc);
    arch_addr_t (*get_pc)(void *dsc);
    void *(*get_bp)(void);
    void (*swap_bp)(void *dsc,void *covered,arch_addr_t addr);
};

#endif /* ARCH_H */
