/*
 * data.h - Values
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#ifndef DATA_H
#define DATA_H

#include "arch.h"
#include "dwarfx.h"


/*
 * Note: a local object can have non-NULL process, e.g. if it's a pointer to an
 * object in a remote process.
 */

#define VF_REMOTE	2	/* data object is stored in remote process */
#define VF_MALLOCED	4	/* free() upon destruction */

/*
 * Field use:
 *
 * VF_REMOTE is only set if the value is stored in a remote process. Its
 *	location is indicated by u.addr. Note that locally stored pointers to
 *	a remote process do not set VF_REMOTE ! Also, process, breakpoint, or
 *	location values do not set VF_REMOTE.
 * "owner" points to the object holding the value in question. There are three
 *	cases where this can happen:
 *	1) the current value is an element of a struct or union
 *	2) the current value is an array element
 *	3) the current value is a locally generated pointer. Note that, if this
 *	   pointer points to a remote process, only "process" of the owner is
 *	   non-NULL. Also note that we can have local pointers where "owner" is
 *	   NULL, i.e. if they have been obtained with "read".
 * "process" provides the process context for the value. If VF_REMOTE is set,
 *	the process context identifies the process in which the value is
 *	stored. If VF_REMOTE is not set, the process context
 *	a) indicates that the object in question does in fact have such a
 *	   context (this distinguishes locally stored pointers pointing to
 *	   local data (void *) from those pointing to remote data
 *	   (arch_addr_t).)
 *	b) provides the context for accesses (and for compatibility checks,
 *	   e.g. in assignments, or when comparing or subtracting pointers)
 * "u.value" points to the value in local memory. VM_MALLOC indicates whether
 *	this memory object should be deallocated upon destruction of the value.
 *      "u.value" is used if VF_REMOTE is not set, and the object type is not
 *	"location_type".
 * "u.addr" contains the address of the object in the remote address space.
 *	"u.addr" is used iff VF_REMOTE is set.
 * "u.obj" points to the dwarfx descriptor of the compilation unit, function,
 *	or label. "u.obj" is used iff the object type is "location_type".
 */

struct value {
    int flags;				/* from VF_* */
    const struct dwx_type *type;	/* type of value (can't be NULL) */
    struct process *process;            /* related process, or NULL */
    struct value *owner;		/* data is inside another object,
					   pointed to by "owner"
					   (only used for dump data ) */
    int ref;				/* reference count */
    union {
	void *value;			/* pointer to value */
	struct {
	    const struct dwx_loc_expr *loc; /* location list */
	    arch_addr_t offset;		/* offset from location */
	} addr;
	const struct dwx_obj *obj;	/* object (for location) */
    } u;
};

#if 0
    /* ----- Remote objects, or data derived from them --------------------- */
    const struct dwx_obj *obj;          /* related object, or NULL */
    const struct dwx_loc_expr *loc;     /* location list */
};

/*
 * Field usage:
 *
 *		value/addr/obj	process
 * process	NULL		pt_process
 * breakpoint	struct breakp.. pt_process
 * location	object		pt_process	(unit/fn/breakpoint)
 * rmt variable	address		pt_process
 */

#endif


#include "op.h"	/* must follow definition of struct value ! */


extern const struct dwx_type *int8_type;
extern const struct dwx_type *int16_type;
extern const struct dwx_type *int32_type;
extern const struct dwx_type *int64_type;
extern const struct dwx_type *uint8_type;
extern const struct dwx_type *uint16_type;
extern const struct dwx_type *uint32_type;
extern const struct dwx_type *uint64_type;

extern struct dwx_type type_type;
extern struct dwx_type string_type;
extern struct dwx_type lvalue_type;
extern struct dwx_type array_type;


int private_type(const struct dwx_type *type);

struct value *new_value(const struct dwx_type *type);
struct value *get(struct value *value);
void put(struct value *value);
/*
 * Note: share does not put its arguments !
 */
void share(struct value *value,struct value *owner);

struct value *int_value(arch_reg_t num);
arch_reg_t value_int(struct value *value);
arch_reg_t int_eval(const struct op *op);

/*
 * If str isn't strdup'ed, it must be constant, and caller needs to bump
 * reference count with get().
 *
 * Warning: use value_string and string_eval only if the string has a reference
 * count > 1 !
 */
struct value *string_value(const char *str,int malloced);
const char *__value_string(struct value *value);
const char *value_string(struct value *value);
const char *string_eval(const struct op *op);

struct value *type_value(const struct dwx_type *type);
const struct dwx_type *value_type(struct value *value);
const struct dwx_type *type_eval(const struct op *op);

struct value *array_value(void);

struct value *lvalue_value(struct value **value);
struct value *evaluate_lvalue(const struct op *op);
struct value *dereference_lvalue(struct value *lvalue);
struct value *evaluate(const struct op *op);

void init_types(void);

#endif /* DATA_H */
