/*
 * process.c - Process and breakpoint handling in debugger
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <sys/wait.h>

#include "arch.h"
#include "dwarfx.h"
#include "ptrace.h"

#include "util.h"
#include "error.h"
#include "data.h"
#include "var.h"
#include "op.h"
#include "stmt.h"
#include "sbug.h"
#include "process.h"


extern void open_elf(const char *name);


static struct process *processes = NULL;


/* ----- Helper functions -------------------------------------------------- */


static int process_has_terminated(const struct process *process)
{
    return process->proc->terminated;
}


void check_process_termination(const struct process *process)
{
    if (!process) clerrorf("referenced process no longer exists");
    if (process_has_terminated(process)) clerrorf("process has terminated");
}


/* ----- Context ----------------------------------------------------------- */


static struct value **current_process_ptr = NULL;
  /* points into $$; NULL if we haven't touched any processes yet  */


static struct process *__current_process(void)
{
    if (!current_process_ptr || !*current_process_ptr ||
      (*current_process_ptr)->type != &process_type)
	return NULL;
    return __value_process(*current_process_ptr);
}


struct process *current_process(void)
{
    struct process *process;

    process = __current_process();
    if (!process) clerrorf("no current process");
    return process;
}


static struct breakpoint *lookup_breakpoint(const struct process *process,
  struct pt_break *bp)
{
    struct breakpoint *b;

    if (!bp) return NULL;
    for (b = process->breakpoints; b; b = b->next)
	if (b->addr == bp->addr) break;
    return b;
}


static void set_current_process(struct process *process)
{
    struct var *var;

    if (!current_process_ptr) {
	var = var_lookup("$");
	current_process_ptr = &var->value;
    }
    if (*current_process_ptr) put(*current_process_ptr);
    *current_process_ptr = process ? get(process->value) : NULL;
    if (process)
	process->current_breakpoint =
	  lookup_breakpoint(process,process->proc->stopped);
}


/* ----- Values ------------------------------------------------------------ */


struct dwx_type process_type;
struct dwx_type breakpoint_type;
struct dwx_type location_type;


struct value *process_value(struct process *process)
{
    struct value *value;

    value = new_value(&process_type);
    value->u.value = process;
    process->value = value;
    return value;
}


struct process *__value_process(struct value *value)
{
    if (!value || value->type != &process_type)
        clerrorf("invalid type (expected process)");
    return value->u.value;
}


struct value *breakpoint_value(struct breakpoint *breakpoint)
{
    struct value *value;

    value = new_value(&breakpoint_type);
    value->u.value = breakpoint;
    breakpoint->value = value;
    return value;
}


struct breakpoint *__value_breakpoint(struct value *value)
{
    if (!value || value->type != &breakpoint_type)
        clerrorf("invalid type (expected breakpoint)");
    return value->u.value;
}


int sbug_put(struct value *value)
{
    if (value->type == &process_type) {
	struct process *process = value->u.value;
	struct process **walk;
	struct breakpoint *bp;

	pt_terminate(process->proc);
	for (walk = &processes; *walk != process; walk = &(*walk)->next);
	*walk = process->next;
	for (bp = process->breakpoints; bp; bp = bp->next) {
	    assert(bp->process == process);
	    bp->process = NULL;
	}
	free(process);
	return 1;
    }
    if (value->type == &breakpoint_type) {
	struct breakpoint *bp = (struct breakpoint *) value->u.value;

	if (bp->process) {
	    struct breakpoint **walk;

	    if (!process_has_terminated(bp->process))
		pt_unbreak(bp->process->proc,bp->addr);
	    for (walk = &bp->process->breakpoints; *walk != bp;
	      walk = &(*walk)->next);
	    *walk = bp->next;
	    if (bp->process->current_breakpoint == bp)
		bp->process->current_breakpoint = NULL;
	}
	free(bp);
	return 1;
    }
    if (value->type == &location_type) return 1;
    return 0;
}


struct process *get_process(struct process *process)
{
    if (process) get(process->value);
    return process;
}


void put_process(struct process *process)
{
    if (process) put(process->value);
}


/* ----- Operators: run ---------------------------------------------------- */


static int need_atexit = 1;


static void kill_processes(void)
{
    struct process *process;

    for (process = processes; process; process = process->next)
	pt_terminate(process->proc);
}


static char *lookup_program(char *name)
{
    const char *path,*pos,*colon;
    int len;

    if (strchr(name,'/')) return stralloc(name);
    path = getenv("PATH");
    if (!path) goto not_found;
    len = strlen(name);
    for (pos = path; *pos; pos = colon+1) {
	const char *end;
	char *tmp;

	colon = strchr(pos,':');
	if (colon) end = colon;
	else {
	    end = strchr(pos,0);
	    colon = end-1;
	}
	tmp = alloc(end-pos+len+2);
	    /* +2: / and \0 */
	memcpy(tmp,pos,end-pos);
	tmp[end-pos] = '/';
	strcpy(tmp+(end-pos)+1,name);
	if (!access(tmp,R_OK | X_OK)) return tmp;
	free(tmp);
    }
    /* fall through */

not_found:
    clerrorf("%s: command not found",name);
}


static struct dwx_obj *load_symbols(const char *path)
{
    struct dwx_obj *objs = NULL,**link = &objs;

    open_elf(path);
    while (1) {
	const void *p;

	p = dwarf_comp_unit();
	if (!p) break;
	*link = dwx_read_objects(p);
	link = &(*link)->next;
    }
    return objs;
}


static struct pt_process *start_program(const char *path,char **argv)
{
    struct pt_process *proc;

    proc = pt_fork();
    if (proc) return proc;
    if (execvp(path,argv) < 0) {
	perror("execvp");
	exit(1);
    }
    /* can't happen */
    exit(0);
}


static struct pt_process *start_uml(const char *path,char **argv)
{
    extern pid_t launch_uml(const char *path,char **argv);

    return pt_new_process(launch_uml(path,argv));
}


static struct process *launch_process(
  struct pt_process *(*start)(const char *path,char **argv),
  const char *path,char **argv)
{
    struct process *process;

    process = alloc_t(struct process);
	/* allocate before fork, so we're sure it works */
    process->proc = start(path,argv);
    process->current_breakpoint = NULL;
    process->breakpoints = NULL;
    process->next = processes;
    processes = process;
    if (need_atexit) {
	atexit(kill_processes);
	need_atexit = 0;
    }
    return process;
}


static struct value *do_op_run(const struct op *self,
  struct pt_process *(*start)(const char *path,char **argv))
{
    struct arg_list *arg;
    int args;
    struct value *value0,*value;
    char **argv;
    char *path;
    struct dwx_obj *objs;
    struct process *process;

    args = 0;
    for (arg = self->u.run_args.args; arg; arg = arg->next)
	args++;
    argv = alloc((args+2)*sizeof(const char *));
	/* +2: argv[0], terminating NULL */
    value0 = evaluate(self->u.run_args.cmd);
    argv[0] = (char *) __value_string(value0);
    args = 1;
    for (arg = self->u.run_args.args; arg; arg = arg->next) {
	assert(!arg->value);
	arg->value = evaluate(arg->expr);
	argv[args] = (char *) __value_string(arg->value);
	args++;
    }
    argv[args] = NULL;
    path = lookup_program(*argv);
    objs = load_symbols(path);
    process = launch_process(start,path,argv);
    process->objs = objs;
    value = process_value(process);
    set_current_process(process);
    free(path);
    for (arg = self->u.run_args.args; arg; arg = arg->next) {
	put(arg->value);
	arg->value = NULL;
    }
    put(value0);
    free(argv);
    return value;
}


struct value *op_run(const struct op *self)
{
    return do_op_run(self,start_program);
}


struct value *op_uml(const struct op *self)
{
    return do_op_run(self,start_uml);
}


/* ----- Operators: break  ------------------------------------------------- */


struct value *op_breakpoint(const struct op *self)
{
    struct value *value;
    struct breakpoint *bp;

    value = evaluate(self->u.arg.a);
    if (!value || value->type != &location_type ||
      value->u.obj->kind == dwx_obj_unit)
	clerrorf("invalid object type (expected function or label)");
    check_process_termination(value->process);
    if (debug)
	fprintf(stderr,"breakpoint at 0x%lx\n",
	  (unsigned long) value->u.obj->low_pc);
    pt_break(value->process->proc,value->u.obj->low_pc);
    bp = alloc_t(struct breakpoint);
    bp->obj = value->u.obj;
    bp->addr = value->u.obj->low_pc;
    bp->process = value->process;
    bp->next = value->process->breakpoints;
    value->process->breakpoints = bp;
    put(value);
    return breakpoint_value(bp);
}


/* ----- Operators: destructor --------------------------------------------- */


int destroy_sbug_op(const struct op *op)
{
    if (op->fn == op_run) {
	destroy_expr(op->u.run_args.cmd);
	destroy_arg_list(op->u.run_args.args);
	return 1;
    }
    return 0;
}


/* ----- Operators: current breakpoint ------------------------------------- */


struct value *op_current_bp(const struct op *self)
{
    struct process *curr;

    curr = __current_process();
    if (!curr || !curr->current_breakpoint) return NULL;
    return get(curr->current_breakpoint->value);
}


/* ----- Operators: current status ----------------------------------------- */


struct value *op_current_status(const struct op *self)
{
    struct process *curr;
    int status;

    curr = __current_process();
    if (!curr || curr->proc->running || curr->proc->stopped) return NULL;
    status = curr->proc->status;
    if (WIFEXITED(status)) status = WEXITSTATUS(status);
    else if (WIFSIGNALED(status)) status = 0x80 | WTERMSIG(status);
    else status = 0;
    return int_value(status);
}


/* ----- Operators: op_eq and op_ne extension ------------------------------ */


static int sbug_relop(struct value *a,struct value *b,struct value **res,
  int not)
{
    int eq;

    if (a && b && a->type != b->type) return 0;
    if ((a && a->type == &process_type) ||
      (!a && b && b->type == &process_type))
	eq = (a ? __value_process(a) : NULL) ==
	  (b ? __value_process(b) : NULL);
    else if ((a && a->type == &breakpoint_type) ||
      (!a && b && b->type == &breakpoint_type))
	eq = (a ? __value_breakpoint(a) : NULL) ==
	  (b ? __value_breakpoint(b) : NULL);
    else return 0;
    *res = int_value(eq ^ not);
    put(a);
    put(b);
    return 1;
}


int sbug_op_eq(struct value *a,struct value *b,struct value **res)
{
    return sbug_relop(a,b,res,0);
}


int sbug_op_ne(struct value *a,struct value *b,struct value **res)
{
    return sbug_relop(a,b,res,1);
}


/* ----- Operators: wait --------------------------------------------------- */


struct value *op_wait(const struct op *self)
{
    struct value *value;
    struct pt_process *proc;
    struct process *process;

    value = self->u.arg.a ? evaluate(self->u.arg.a) : NULL;
    process = value ? __value_process(value) : NULL;
    if (process) {
	if (process->proc->running != 1)
	    clerrorf("process must be running");
    }
    else {
	const struct process *walk;

	for (walk = processes; walk; walk = walk->next)
	    if (walk->proc->running) break;
	if (!walk) clerrorf("no running processes");
    }
    proc = pt_wait(process ? process->proc : NULL,1);
    put(value);
    for (process = processes; process->proc != proc; process = process->next);
    set_current_process(process);
    return NULL; /* @@@ don't know useful result yet */
}


/* ----- Statements -------------------------------------------------------- */


void stmt_continue_process(const struct stmt *self)
{
    struct value *value;
    struct process *process;

    if (!self->u.process.process) process = current_process();
    else {
	value = evaluate(self->u.process.process);
	process = __value_process(value);
    }
    set_current_process(NULL);
    check_process_termination(process);
    pt_run(process->proc);
    if (self->u.process.process) put(value);
}


int destroy_sbug_stmt(const struct stmt *stmt)
{
    if (stmt->fn == stmt_continue_process) {
	destroy_expr(stmt->u.process.process);
	return 1;
    }
    return 0;
}
