/*
 * ptrace.c - Ptrace wrapper
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

/*
 * The design of the breakpoint handling mechanism is taken from kprobes:
 * http://www-124.ibm.com/linux/projects/kprobes/
 */

/*
 * @@@ Known bug:
 * Pointer/address size may be different in tracer and in program being traced.
 * E.g. on sparc64, user space is 32 bit, but kernel is 64 bit. So If we're
 * tracing a correct UML representation of the kernel, address sizes will be
 * different.
 *
 * Note that alignment is handled correctly, i.e. all accesses to the traced
 * process are aligned to a multiple of the host word size.
 *
 * Note that we assume the traced process to be relatively cooperative. We
 * don't try to prevent it from exec'ing something else, or from changing its
 * code (including any breakpoints).
 */

/*
 * Old versions of the ptrace man page contain the following errors:
 *
 * - the "data" argument of PTRACE_POKETEXT etc. is the value, not its address
 * - (PTRACE_ATTACH) the parent PID is returned by getppid, not getpid
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/wait.h>

#include "util.h"
#include "arch.h"
#include "ptrace.h"


extern struct arch_ops ia32_arch_ops;


static struct pt_process *procs = NULL;


static void ptrace_read(pid_t pid,int peek_op,arch_addr_t addr,
  void *buf,size_t size);
static void ptrace_write(pid_t pid,int peek_op,int poke_op,arch_addr_t addr,
  const void *buf,size_t size);


/* ----- Data I/O ---------------------------------------------------------- */


static long ptrace_peek_data(pid_t pid,int peek_op,void *addr)
{
    long word;

    word = ptrace(peek_op,pid,addr,NULL);
    if (word == -1 && errno) {
	perror("ptrace(PTRACE_PEEK...)");
	exit(1);
    }
    return word;
}


void pt_swap_byte(struct pt_process *proc,arch_addr_t addr,uint8_t *byte)
{
    unsigned long a = addr & ~(sizeof(long)-1);
    int off = addr-a;
    long word;
    uint8_t tmp;

    errno = 0;
    word = ptrace_peek_data(proc->pid,PTRACE_PEEKDATA,(void *) a);
    tmp = word >> (off*8);
    word &= ~(0xff << (off*8));
    word |= *byte << (off*8);
    if (ptrace(PTRACE_POKEDATA,proc->pid,(void *) a,(long) word) < 0) {
	perror("ptrace(PTRACE_POKEDATA)");
	exit(1);
    }
    *byte = tmp;
}


static void ptrace_read(pid_t pid,int peek_op,arch_addr_t addr,
  void *buf,size_t size)
{
    unsigned long a = addr & ~(sizeof(long)-1);
    int off = addr-a;
    long word;

    if (a != addr) {
	int len;

	word = ptrace_peek_data(pid,peek_op,(void *) a);
	len = sizeof(long)-off;
	if (len > size) len = size;
	memcpy(buf,((uint8_t *) &word)+off,len);
	a += sizeof(long);
	buf += len;
	size -= len;
    }
    while (size >= sizeof(long)) {
	word = ptrace_peek_data(pid,peek_op,(void *) a);
	*(long *) buf = word;
	a += sizeof(long);
	buf += sizeof(long);
	size -= sizeof(long);
    }
    if (size) {
	word = ptrace_peek_data(pid,peek_op,(void *) a);
	memcpy(buf,&word,size);
    }
}


void pt_read(struct pt_process *proc,arch_addr_t addr,void *buf,size_t size)
{
    assert(proc->running != 1);
    ptrace_read(proc->pid,PTRACE_PEEKDATA,addr,buf,size);
}


static void ptrace_write(pid_t pid,int peek_op,int poke_op,arch_addr_t addr,
  const void *buf,size_t size)
{
    unsigned long a = addr & ~(sizeof(long)-1);
    int off = addr-a;
    long word;

    if (a != addr) {
	int len;

	word = ptrace_peek_data(pid,peek_op,(void *) a);
	len = sizeof(long)-off;
	if (len > size) len = size;
	memcpy(((uint8_t *) &word)+off,buf,len);
	if (ptrace(poke_op,pid,(void *) a,word) < 0) {
	    perror("ptrace(PTRACE_POKE...)");
	    exit(1);
	}
	a += sizeof(long);
	buf += len;
	size -= len;
    }
    while (size >= sizeof(long)) {
	if (ptrace(poke_op,pid,(void *) a,*(long *) buf) < 0) {
	    perror("ptrace(PTRACE_POKE...)");
	    exit(1);
	}
	a += sizeof(long);
	buf += sizeof(long);
	size -= sizeof(long);
    }
    if (size) {
	word = ptrace_peek_data(pid,peek_op,(void *) a);
	memcpy(&word,buf,size);
	if (ptrace(poke_op,pid,(void *) a,word) < 0) {
	    perror("ptrace(PTRACE_POKE...)");
	    exit(1);
	}
    }
}


void pt_write(struct pt_process *proc,arch_addr_t addr,const void *buf,
  size_t size)
{
    assert(proc->running != 1);
    ptrace_write(proc->pid,PTRACE_PEEKDATA,PTRACE_POKEDATA,addr,buf,size);
}


/* ----- Process creation -------------------------------------------------- */


static void wait_for_signal(pid_t pid,int sig)
{
    int status;

    if (waitpid(pid,&status,WUNTRACED) < 0) {
	perror("waitpid");
	exit(1);
    }
    if (!WIFSTOPPED(status) || WSTOPSIG(status) != sig) {
	fprintf(stderr,"confusing status %d (want %d)\n",status,sig);
	exit(1);
    }
}


struct pt_process *pt_new_process(pid_t pid)
{
    struct pt_process *proc;

    proc = alloc_t(struct pt_process);
    proc->pid = pid;
    proc->running = -1;
    proc->terminated = 0;
    proc->bps = proc->stopped = NULL;
    proc->next = procs;
    procs = proc;
    proc->arch_ops = &ia32_arch_ops;
    proc->arch_dsc = proc->arch_ops->init_arch(proc);
    /*
     * E.g. is we're on a split 32/64 bit architecture, ptrave wouldn't be able
     * to reach all addresses in a 64 bit process if we've been built for 32
     * bit. (Need to use ptrace64 in such cases.)
     */
    if (sizeof(arch_addr_t) > sizeof(unsigned long))
	fprintf(stderr,"warning: "
	  "sizeof(arch_addr_t) > sizeof(unsigned long)\n");
    return proc;
}


struct pt_process *pt_trace(pid_t pid)
{
    if (ptrace(PTRACE_ATTACH,pid,NULL,NULL) < 0) {
	perror("ptrace(PTRACE_ATTACH)");
	exit(1);
    }
    wait_for_signal(pid,SIGSTOP);
    return pt_new_process(pid);
}


struct pt_process *pt_fork(void)
{
    pid_t pid;

    pid = fork();
    if (pid < 0) {
	perror("fork");
	exit(1);
    }
    if (!pid) {
	if (ptrace(PTRACE_TRACEME,0,NULL,NULL) < 0) {
	    perror("ptrace(PTRACE_TRACEME)");
	    exit(1);
	}
	return NULL;
    }
    wait_for_signal(pid,SIGTRAP); /* enter exec */
    if (ptrace(PTRACE_SYSCALL,pid,NULL,NULL) < 0) {
	perror("ptrace(PTRACE_SYSCALL)");
	exit(1);
    }
    wait_for_signal(pid,SIGTRAP); /* leave exec */
    return pt_new_process(pid);
}


/* ----- Process termination ----------------------------------------------- */


void pt_terminate(struct pt_process *proc)
{
    struct pt_process **walk;

    if (!proc->terminated) {
	if (ptrace(PTRACE_KILL,proc->pid,NULL,NULL) < 0) {
	    perror("ptrace(PTRACE_KILL)");
	    exit(1);
	}
	while (1) {
	    if (waitpid(proc->pid,&proc->status,0) < 0) {
		perror("waitpid");
		exit(1);
	    }
	    if (!WIFSTOPPED(proc->status)) break;
	    if (ptrace(PTRACE_CONT,proc->pid,NULL,SIGKILL) < 0) {
		perror("ptrace(PTRACE_CONT)");
		exit(1);
	    }
	}
    }
    assert(WIFEXITED(proc->status) || WIFSIGNALED(proc->status));
    for (walk = &procs; *walk != proc; walk = &(*walk)->next);
    *walk = proc->next;
    while (proc->bps) {
	struct pt_break *next = proc->bps->next;

	free(proc->bps->arch_covered);
	free(proc->bps);
	proc->bps = next;
    }
    free(proc->arch_dsc);
    free(proc);
}


/* ----- Breakpoints ------------------------------------------------------- */


static void pt_step_bp(struct pt_process *proc)
{
    int status;

    proc->arch_ops->swap_bp(proc->arch_dsc,proc->stopped->arch_covered,
      proc->stopped->addr);
    if (ptrace(PTRACE_SINGLESTEP,proc->pid,NULL,NULL) < 0) {
	perror("ptrace(PTRACE_SINGLESTEP)");
	exit(1);
    }
    /*
     * @@@ Not entirely correct: we may get some signal in this step.
     */
    if (waitpid(proc->pid,&status,WUNTRACED) < 0) {
	perror("waitpid");
	exit(1);
    }
    if (!WIFSTOPPED(status) || WSTOPSIG(status) != SIGTRAP) {
	fprintf(stderr,"confusing status %d\n",status);
	exit(1);
    }
    proc->arch_ops->swap_bp(proc->arch_dsc,proc->stopped->arch_covered,
      proc->stopped->addr);
    proc->stopped = NULL;
}


void pt_run(struct pt_process *proc)
{
    assert(proc->running != 1);
    proc->arch_ops->resume(proc->arch_dsc);
    if (proc->stopped) pt_step_bp(proc);
    if (ptrace(PTRACE_CONT,proc->pid,NULL,NULL) < 0) {
	perror("ptrace(PTRACE_CONT)");
	exit(1);
    }
    proc->running = 1;
}


void pt_break(struct pt_process *proc,arch_addr_t addr)
{
    struct pt_break *bp;

    assert(proc->running != 1);
    for (bp = proc->bps; bp; bp = bp->next)
	if (bp->addr == addr) {
	    bp->ref++;
	    return;
	}
    bp = alloc_t(struct pt_break);
    bp->addr = addr;
    bp->arch_covered = proc->arch_ops->get_bp();
    bp->ref = 1;
    bp->next = proc->bps;
    proc->bps = bp;
    proc->arch_ops->swap_bp(proc->arch_dsc,bp->arch_covered,addr);
}


void pt_unbreak(struct pt_process *proc,arch_addr_t addr)
{
    struct pt_break **walk,*bp;

/*
 * @@@ UGLY HACK ALERT !  During cleanup, sbug will deallocate variables in an
 * arbitrary order, so breakpoints may get removed for processes that are still
 * running (or that haven't been reaped yet).
 *
 * Luckily, it seems that we don't need to stop the process to scribble over
 * it, so just not checking that it's not stopped does the trick, in some ugly
 * way.
 *
 * Possible fixes:
 * - remove breakpoints only explicitly, i.e. re-introduce the "delete"
 *   command. That way, automatic breakpoint removal happens only at process
 *   destruction, which we already handle correctly.
 * - SIGSTOP the process, fetch any pending status, remove the breakpoint,
 *   if the status wasn't from the SIGSTOP, resume to fetch SIGSTOP (if
 *   applicable), then write previously fetched status back to descriptor.
 *   Messy.
 * - there must be some other way ... :-)
 */
//    assert(proc->running != 1);
    for (walk = &proc->bps; *walk && (*walk)->addr != addr;
      walk = &(*walk)->next);
    if (!*walk) {
	fprintf(stderr,"no breakpoint at 0x%lx\n",(unsigned long) addr);
	exit(1);
    }
    bp = *walk;
    if (--bp->ref) return;
    *walk = bp->next;
    proc->arch_ops->swap_bp(proc->arch_dsc,bp->arch_covered,bp->addr);
    if (proc->stopped == bp) proc->stopped = NULL;
    free(bp->arch_covered);
    free(bp);
}


/* ----- Dispatch events --------------------------------------------------- */


struct pt_process *pt_wait(const struct pt_process *wait_proc,int block)
{
    while (1) {
	int status;
	pid_t pid;
	struct pt_process *proc;
	struct pt_break *bp;
	arch_addr_t addr;

	if (wait_proc && wait_proc->running != 1) {
	    fprintf(stderr,"pt_wait: process isn't running\n");
	    exit(1);
	}
	pid = waitpid(wait_proc ? wait_proc->pid : -1,&status,
	  (block ? 0 : WNOHANG) | WUNTRACED);
	if (!pid) return NULL;
	if (pid < 0) {
	    perror("waitpid");
	    exit(1);
	}
	if (WIFSTOPPED(status) && WSTOPSIG(status) != SIGTRAP) {
	    if (ptrace(PTRACE_CONT,pid,NULL,WSTOPSIG(status)) < 0) {
		perror("ptrace(PTRACE_CONT)");
		exit(1);
	    }
	    continue;
	}
	for (proc = procs; proc; proc = proc->next)
	    if (proc->pid == pid) break;
	if (!proc) {
	    fprintf(stderr,"process %d not found\n",(int) pid);
	    exit(1);
	}
	proc->running = 0;
	if (WIFEXITED(status) || WIFSIGNALED(status)) {
	    proc->terminated = 1;
	    proc->status = status;
	    return proc;
 	}
	assert(WIFSTOPPED(status));
	proc->arch_ops->stopped(proc->arch_dsc);
	proc->arch_ops->backup_pc(proc->arch_dsc);
	addr = proc->arch_ops->get_pc(proc->arch_dsc);
	for (bp = proc->bps; bp; bp = bp->next)
	    if (bp->addr == addr) break;
	if (!bp) {
	    fprintf(stderr,"no breakpoint for %d:0x%lx\n",(int) pid,
	      (unsigned long) addr);
	    exit(1);
	}
	proc->stopped = bp;
	return proc;
    }
}
