/*
 * uml.c - Creation of, and synchronization with the UML process
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

/*
 * launch_uml starts an instance of UML and synchronizes ptrace attachment with
 * the UML process.
 *
 * When running under a debugger, UML ptraces that debugger, and redirects
 * calls to ptrace, waitpid, etc. to the corresponding child process, which
 * executes the actual UML kernel code.
 *
 * The debugger must wait until UML has attached itself to it, and then it can
 * in turn attach itself to the supposed UML process (PID 1). Unfortunately,
 * UML doesn't tell us when it has attached itself to the debugger. 
 *
 * So we ptrace the UML process to see when it has done this. Things get a bit
 * murky now, e.g. if two processes try to ptrace each other, they may become
 * unkillable, etc. Also, this doesn't really seem to work anyway.
 *
 * The solution is to involve a third process, which ptraces UML, and tells
 * the debugger when UML has attached to the debugger. Then, the debugger can
 * attach to PID 1, which UML, ptracing that debugger, will catch, and
 * redirect.
 *
 * Nice and easy :-)
 */

/*
 * @@@ Should abstract ia-32 code.
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/wait.h>
#include <asm/unistd.h>

#include "util.h"


extern int debug;


#define CHECK(call) \
    do { \
	if ((call) < 0) { \
	    perror(#call); \
	    exit(1); \
	}  \
    } \
    while (0)


static int wait_for_signal(pid_t pid,int sig,int can_stop)
{
    int status;

    CHECK(waitpid(pid,&status,WUNTRACED));
    if (WIFSTOPPED(status)) {
	if (WSTOPSIG(status) == sig) return 0;
	if (can_stop) return WSTOPSIG(status);
	if (debug) fprintf(stderr,"got stop signal %d\n",WSTOPSIG(status));
	exit(1);
    }
    if (WIFSIGNALED(status)) {
	if (debug) fprintf(stderr,"got terminal signal %d\n",WTERMSIG(status));
	exit(1);
    }
    if (WIFEXITED(status)) {
	if (debug) fprintf(stderr,"exit status %d\n",WEXITSTATUS(status));
	exit(1);
    }
    if (debug) fprintf(stderr,"confusing status %d (want %d)\n",status,sig);
    exit(1);
}


static void watcher(const char *path,char **argv,int uml_ready,int main_dead)
{
    char dummy;
    pid_t pid;

    pid = fork();
    if (!pid) {
	CHECK(ptrace(PTRACE_TRACEME,0,NULL,NULL));
	CHECK(execvp(path,argv));
	/* can't happen */
	exit(0);
    }
    if (pid < 0) {
	perror("fork");
	exit(1);
    }
    wait_for_signal(pid,SIGTRAP,0);
    while (1) {
	struct user_regs_struct regs;
	int sig;

	sig = 0;
	do {
	    CHECK(ptrace(PTRACE_SYSCALL,pid,NULL,sig));
	    sig = wait_for_signal(pid,SIGTRAP,1);
	    if (sig && debug) fprintf(stderr,"[signal %d]\n",sig);
	}
	while (sig);
	CHECK(ptrace(PTRACE_GETREGS,pid,NULL,&regs));
	if (debug)
	    fprintf(stderr,"[syscall %u@0x%lx]\n",(unsigned) regs.orig_eax,
	      (unsigned long) regs.eip);
	sig = 0;
	do {
	    CHECK(ptrace(PTRACE_SYSCALL,pid,NULL,sig));
	    sig = wait_for_signal(pid,SIGTRAP,1);
	    if (sig && debug) fprintf(stderr,"[signal %d]\n",sig);
	}
	while (sig);
	if (debug) fprintf(stderr,"[syscall return]\n");
	if (regs.orig_eax == __NR_ptrace) {
	    if (debug) fprintf(stderr,"PTRACE_%d\n",(int) regs.ebx);
	    /*
	     * Assumption: if PTRACE_ATTACH returns, we're all set. If this
	     * isn't true, we have to wait for some additional event. It seems
	     * that waiting for PTRACE_SYSCALL is a good choice.
	     */
	    if (regs.ebx == PTRACE_ATTACH) break;
	}
    }
    CHECK(ptrace(PTRACE_DETACH,pid,NULL,NULL));
    if (debug) fprintf(stderr,"detached\n");
    CHECK(write(uml_ready,"",1)); /* should also check for 0 */
    /*
     * UML will get very confused if the debugger receives a SIGCHLD out of the
     * blue. So we hang around till the debugger dies. (Which implicitly closes
     * our fd.)
     */
    CHECK(read(main_dead,&dummy,1)); /* should also check for 1 as error */
    if (debug) fprintf(stderr,"done\n");
    exit(0);
}


static char **build_argv(char **argv,pid_t main_pid)
{
    char buf[100]; /* enough for 290 bit PIDs :-) */
    char **arg;
    int args = 0;
    char **new_argv;

    for (arg = argv; *arg; arg++) args++;
    new_argv = alloc(sizeof(char *)*(args+3));
	/* +1 for NULL, +1 for debug, +1 for gdb-pid=N */
    new_argv[0] = argv[0];
    sprintf(buf,"gdb-pid=%d",getpid());
    new_argv[1] = buf;
    new_argv[2] = "debug";
    args = 3;
    for (arg = argv+1; *arg; arg++)
	new_argv[args++] = *arg;
    new_argv[args] = NULL;
    return new_argv;
}


pid_t launch_uml(const char *path,char **argv)
{
    int uml_ready[2],main_dead[2];
    char **new_argv;
    pid_t pid;
    char dummy;
    int got;

    CHECK(pipe(uml_ready));
    CHECK(pipe(main_dead));
    new_argv = build_argv(argv,getpid());
    pid = fork();
    if (!pid) {
	CHECK(close(main_dead[1]));
	CHECK(close(uml_ready[0]));
	watcher(path,new_argv,uml_ready[1],main_dead[0]);
    }
    if (pid < 0) {
	perror("fork");
	exit(1);
    }
    free(new_argv);
    CHECK(close(main_dead[0]));
    CHECK(close(uml_ready[1]));
    got = read(uml_ready[0],&dummy,1);
    if (got < 0) {
	perror("read");
	exit(1);
    }
    if (!got) {
	fprintf(stderr,"Watcher died. - Terminating.\n");
	exit(1);
    }
    CHECK(ptrace(PTRACE_ATTACH,1,NULL,NULL));
    if (debug) fprintf(stderr,"re-attached\n");
    (void) wait_for_signal(1,SIGSTOP,0);
    if (debug) fprintf(stderr,"stopped, ready to go !\n");
    return 1; /* UML's pseudo pid */
}
