/*
 * dwxloc.c - Location expressions
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "arch.h"
#include "util.h"

#include "dwarf.h"
#include "dwarfhelp.h"
#include "dwxloc.h"


#define STACK_SIZE 1024


extern int debug;


/* ----- Compute locations ------------------------------------------------- */


static struct dwx_loc stack[STACK_SIZE];
static int sp;


static const struct arch_ops *arch_ops;
static void *arch_dsc;
static const struct dwx_loc_expr *frame_base;


void dwx_loc_setup_hack(const struct arch_ops *ops,void *dsc,
  const struct dwx_loc_expr *fb)
{
    arch_ops = ops;
    arch_dsc = dsc;
    frame_base = fb;
}


static inline struct dwx_loc addr(arch_reg_t v)
{
    struct dwx_loc loc;

    loc.reg = -1;
    loc.addr = v;
    return loc;
}


static inline struct dwx_loc reg(int num)
{
    struct dwx_loc loc;

    loc.reg = num;
    return loc;
}


static inline arch_reg_t eval(struct dwx_loc loc)
{
    if (loc.reg == -1) return loc.addr;
    return arch_ops->get_reg(arch_dsc,loc.reg);
}


static inline arch_addr_t dereference(arch_addr_t addr,int size)
{
    abort();
}


#define RELOP(op) \
    assert(sp > 0); \
    stack[sp-1] = addr((arch_sreg_t) eval(stack[sp]) op \
      (arch_sreg_t) eval(stack[sp-1])); \
    sp--


static void run_op(const void **pc)
{
    enum DW_OP op;
    struct dwx_loc tmp;
    arch_addr_t tmp_addr;
    const void *tmp_pc;

    op = *(unsigned char *) *pc;
    if (debug) printf("op 0x%x, pc 0x%p, sp %d\n",op,pc,sp);
    (*pc)++;
    switch (op) {
	case DW_OP_addr:
	    stack[++sp] = addr(read_addr(*pc));
	    *pc += 4; /* @@@ */
	    break;
	case DW_OP_deref:
	    assert(sp >= 0);
	    stack[sp] = addr(dereference(eval(stack[sp]),sizeof(arch_addr_t)));
	    break;
	case DW_OP_const1s:
	    stack[++sp] = addr(
	      signex_tn(arch_addr_t,*(unsigned char *) *pc,1));
	    (*pc)++;
	    break;
	case DW_OP_const1u:
	    stack[++sp] = addr(*(unsigned char *) *pc);
	    (*pc)++;
	    break;
	case DW_OP_const2s:
	    stack[++sp] = addr(
	      signex_tn(arch_addr_t,read_16(*pc),2));
	    *pc += 2;
	    break;
	case DW_OP_const2u:
	    stack[++sp] = addr(read_16(*pc));
	    *pc += 2;
	    break;
	case DW_OP_const4s:
	    stack[++sp] = addr(
	      signex_tn(arch_addr_t,read_32(*pc),4));
	    *pc += 4;
	    break;
	case DW_OP_const4u:
	    stack[++sp] = addr(read_32(*pc));
	    *pc += 4;
	    break;
/* @@@	case DW_OP_const8s: */
/* @@@	case DW_OP_const8u: */
	case DW_OP_consts:
	    stack[++sp] = addr(sLEB128_decode(pc));
	    break;
	case DW_OP_constu:
	    stack[++sp] = addr(uLEB128_decode(pc));
	    break;
	case DW_OP_dup:
	    assert(sp >= 0);
	    stack[sp+1] = stack[sp];
	    sp++;
	    break;
	case DW_OP_drop:
	    assert(sp >= 0);
	    sp--;
	    break;
	case DW_OP_over:
	    assert(sp > -1);
	    stack[sp+1] = stack[sp-1];
	    sp++;
	    break;
	case DW_OP_pick:
	    assert(sp >= *(unsigned char *) *pc);
	    stack[sp+1] = stack[sp-*(unsigned char *) *pc];
	    (*pc)++;
	    sp++;
	    break;
	case DW_OP_swap:
	    assert(sp > -1);
	    tmp = stack[sp];
	    stack[sp] = stack[sp-1];
	    stack[sp-1] = tmp;
	    break;
	case DW_OP_rot:
	    assert(sp > 2);
	    tmp = stack[sp];
	    stack[sp] = stack[sp-1];
	    stack[sp-1] = stack[sp-2];
	    stack[sp-2] = tmp;
	    break;
	case DW_OP_xderef:
	    abort();
	case DW_OP_abs:
	    assert(sp > -1);
	    tmp_addr = eval(stack[sp]);
	    if ((arch_sreg_t) tmp_addr < 0) stack[sp] = addr(-tmp_addr);
	    break;
	case DW_OP_and:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) & eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_div:
	    assert(sp > 0);
	    stack[sp-1] = addr((arch_sreg_t) eval(stack[sp-1]) /
	      (arch_sreg_t) eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_minus:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) - eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_mod:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) % eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_mul:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) * eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_neg:
	    assert(sp >= 0);
	    stack[sp] = addr(-(arch_sreg_t) eval(stack[sp]));
	    break;
	case DW_OP_not:
	    assert(sp >= 0);
	    stack[sp] = addr(~eval(stack[sp]));
	    break;
	case DW_OP_or:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) | eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_plus:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) + eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_plus_uconst:
	    assert(sp >= 0);
	    stack[sp] = addr(eval(stack[sp])+uLEB128_decode(pc));
	    break;
	case DW_OP_shl:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) << eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_shr:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) >> eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_shra:
	    assert(sp > 0);
	    tmp_addr = eval(stack[sp-1]);
	    if ((arch_sreg_t) tmp_addr >= 0)
		stack[sp-1] = addr(tmp_addr >> eval(stack[sp]));
	    else stack[sp-1] = addr(-(arch_sreg_t)
		  (-(arch_reg_t) tmp_addr >> eval(stack[sp])));
	    sp--;
	    break;
	case DW_OP_xor:
	    assert(sp > 0);
	    stack[sp-1] = addr(eval(stack[sp-1]) ^ eval(stack[sp]));
	    sp--;
	    break;
	case DW_OP_skip:
	    pc += read_16(*pc)+2;
	    break;
	case DW_OP_bra:
	    assert(sp > 0);
	    pc += (eval(stack[sp]) ? read_16(*pc) : 0)+2;
	    sp--;
	    break;
	case DW_OP_eq:
	    RELOP(==);
	    break;
	case DW_OP_ge:
	    RELOP(>=);
	    break;
	case DW_OP_gt:
	    RELOP(>);
	    break;
	case DW_OP_le:
	    RELOP(<=);
	    break;
	case DW_OP_lt:
	    RELOP(<);
	    break;
	case DW_OP_ne:
	    RELOP(!=);
	    break;
	case DW_OP_regx:
	    stack[++sp].reg = uLEB128_decode(pc);
	    abort();
	case DW_OP_fbreg:
	    tmp_pc = frame_base->expr;
	    while (tmp_pc != frame_base->expr+frame_base->length)
		run_op(&tmp_pc);
	    assert(sp >= 0);
	    stack[sp] = addr(eval(stack[sp])+sLEB128_decode(pc));
	    break;
	case DW_OP_bregx:
	    /* see DW_OP_bregN, below. leave abort for now to catch real
	       encoding example */
	    abort();
	case DW_OP_piece:
	    abort();
	case DW_OP_deref_size:
	    assert(sp >= 0);
	    stack[sp] = addr(dereference(eval(stack[sp]),
	      *(unsigned char *) *pc));
	    (*pc)++;
	    break;
	case DW_OP_xderef_size:
	    abort();
	case DW_OP_nop:
	    break;
	case DW_OP_lo_user:
	    abort();
	case DW_OP_hi_user:
	    abort();
	default:
	    if (op >= DW_OP_lit0 && op <= DW_OP_lit31) {
		stack[++sp] = addr(op-DW_OP_lit0);
		break;
	    }
	    if (op >= DW_OP_reg0 && op <= DW_OP_reg31) {
		stack[++sp].reg = op-DW_OP_reg0;
		break;
	    }
	    if (op >= DW_OP_breg0 && op <= DW_OP_breg31) {
		stack[++sp] = addr(eval(reg(op-DW_OP_reg0))+
		  sLEB128_decode(pc));
		break;
	    }
	    fprintf(stderr,"unrecognized opcode %u\n",op);
	    exit(1);
    }
}


arch_addr_t dwx_loc_comp(const void *expr,int length,arch_addr_t base)
{
    const void *end = expr+length;

    stack[0] = addr(base);
    sp = 0;
    while (expr != end)
	run_op(&expr);
    assert(!sp);
    assert(stack[sp].reg == -1);
    return eval(stack[sp]);
}


struct dwx_loc dwx_loc_1(const struct dwx_loc_expr *loc)
{
    const void *expr = loc->expr;
    const void *end = expr+loc->length;

    sp = -1;
    while (expr != end)
	run_op(&expr);
    assert(!sp);
    return stack[sp];
}


const struct dwx_loc_expr *dwx_loc_expr_here(const struct dwx_loc_expr *list,
  arch_addr_t addr)
{
    const struct dwx_loc_expr *walk;

    for (walk = list; walk; walk = walk->next)
	if (walk->start <= addr && walk->end > addr) break;
    return walk;
}


struct dwx_loc dwx_loc_here(const struct dwx_loc_expr *list,arch_addr_t addr)
{
    const struct dwx_loc_expr *expr;

    expr = dwx_loc_expr_here(list,addr);
    if (expr) return dwx_loc_1(expr);
    fprintf(stderr,"location unknown for address 0x%lx\n",
      (unsigned long) addr);
    exit(1);
}


/* ----- Read location list ------------------------------------------------ */


const struct dwx_loc_expr *dwx_loc_list(enum DW_FORM form,const void *pos)
{
    struct dwx_loc_expr *loc;
    const void *ref;

    switch (form) {
	case DW_FORM_block1:
	case DW_FORM_block2:
	case DW_FORM_block4:
	case DW_FORM_block:
	    loc = alloc_t(struct dwx_loc_expr);
	    loc->start = 0;
	    loc->end = ~0;
	    loc->expr = dwarf_decode_block(form,pos,&loc->length);
	    loc->next = NULL;
	    return loc;
	default:
	    break;
    }
    ref = dwarf_decode_reference(form,pos);
    loc = NULL;
    while (1) {
	struct dwx_loc_expr *new;

	new = alloc_t(struct dwx_loc_expr);
	new->start = read_32(ref);
	ref += 4;
	new->end = read_32(ref);
	ref += 4;
	if (!new->start && !new->end) {
	    free(new);
	    break;
	}
	new->length = read_16(ref);
	ref += 2;
	new->expr = ref;
	ref += new->length;
	new->next = loc;
	loc = new;
    }
    return loc;
}
